unit umain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Parser, ExtCtrls, ShellApi;

const
 CFI : Integer = 0;
 GFH : Integer = 1;
 GFI : Integer = 2;

type
  TForm1 = class(TForm)
    Memo1: TMemo;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    Braces : Integer;
    Comments : Integer;
    LineComment, define, hashif, hashelse, hashendif, include : boolean;
    ClassHeaders, StructHeaders, FunctionHeaders, Functions, ClassFunctions : TStrings;
    procedure ParseFile(filename : String);
    function CommentOut(var text : String; StartString, EndString : PChar; var NestLevel : Integer) : boolean;
    function GetString(text : PChar) : String;
    function LineCommentOut(var text : String; StartString, EndString : String; var inblock : boolean) : boolean;
    function GetStringToColon(text : PChar) : String;
    function GetStringtoBracket(text : PChar) : String;
    procedure StartParsingFile;
    procedure FinishParsingFile;
    procedure ParseRecord(text : String);
    function FormatFunction(text : PChar; var output : String) : Integer;
    function Compare(text : string; index : integer; startstring : string) : boolean;
    function RemoveLineFeeds(text : PChar) : string;
    procedure WMDROPFILES(var Message: TWMDROPFILES);
      message WM_DROPFILES;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.ParseRecord(text : String);
var
 temp : String;
 pc : pchar;
 ftype : Integer;
begin

 // check to see if we are within any blocks which we ignore
 LineCommentOut(text, '//', #13#10, linecomment);
 CommentOut(text, '/*', '*/', Comments);
 LineCommentOut(text, '#define', #13#10, define);
 LineCommentOut(text, '#if', #13#10, hashif);
 LineCommentOut(text, '#el', #13#10, hashelse);
 LineCommentOut(text, '#endif', #13#10, hashendif);
 LineCommentOut(text, '#include', #13#10, include);
 CommentOut(text, '{', '}', Braces);

 Trim(text);
 // find any class headers
 pc := strpos(pchar(text), 'class ');
 if pc <> Nil then begin
  ClassHeaders.add('3. ' + RemoveLineFeeds(pc));
  exit;
 end;

 // find any struct headers
 pc := strpos(pchar(text), 'struct ');
 if pc <> Nil then begin
  StructHeaders.add('3. ' + RemoveLineFeeds(pc));
  exit
 end;

 // find any function headers or definitions
 pc := strpos(pchar(text), '(');
 if pc <> Nil then begin
  text := Trim(text);
  text := RemoveLineFeeds(PChar(Text));
  ftype := FormatFunction(PChar(text), temp);
  if ftype = GFH then FunctionHeaders.add('3. ' + temp);
  if ftype = GFI then Functions.add('3. ' + temp);
  if ftype = CFI then ClassFunctions.add('3. ' + temp);
 end;
end;

// this function returns true if the text is currently within a block to be ignored
// currently these are comment blocks and braces, it takes account of nesting
// It does this by taking an input field and removing any start end pairs
// completely. Then it counts start and end pairs and increments / decrements
// the nesting level accordingly. If the nesting level is greater than zero then
// the text should be ignored.
function TForm1.CommentOut(var text : String; StartString, EndString : PChar; var NestLevel : Integer) : boolean;
var
 i, j : Integer;
begin
 for i := 1 to Length(text) do
 begin
  if Compare(text, i, startstring) = true then begin
   inc(NestLevel);
   for j := 0 to Length(Startstring) - 1 do text[i + j] := ' ';
  end;
  if Compare(text, i, endstring) = true then begin
   dec(NestLevel);
   for j := 0 to Length(endstring) - 1 do text[i + j] := ' ';
  end;
  if Nestlevel > 0 then text[i] := ' ';
 end;

end;

// this does not allow for nested
function TForm1.LineCommentOut(var text : String; StartString, EndString : String; var inblock : boolean) : boolean;
var
 i, pc, pc2 : integer;
begin
 // if currently in a block check to see if an end character has come along
 // if an end character has not come along we must remove all the text
 if inblock = true then begin
  pc := pos(EndString, text);
  if pc > 0 then begin
   inblock := false;
   for i := 1 to pc + length(EndString) - 1 do text[i] := #32;
  end
  else
   for i := 1 to length(text) do text[i] := #32;
 end;

 // remove all start / end pairs in text
 pc := pos(StartString, text);
 repeat
  if pc > 0 then begin
   pc2 := pos(EndString, Copy(text, pc  + Length(StartString), Length(text)));
   if pc2 > 0 then begin
    pc2 := pc2 + pc + length(EndString) + Length(StartString) - 2;
    for i :=  pc to pc2 do text[i] := #32;
   end
  end;
  pc := pos(StartString, text);
 until (pc = 0) or (pc2 = 0);

 // if there is a starter without a finisher remove all remaining text
 pc := pos(StartString, text);
 if (pc > 0) then begin
  inblock := true;
   for i := pc to Length(text) do text[i] := #32;
 end;

 result := inblock;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
 ClassHeaders := TStringList.Create;
 StructHeaders := TStringList.Create;
 FunctionHeaders := TStringList.Create;
 Functions := TStringList.Create;
 ClassFunctions := TStringList.Create;
 DragAcceptFiles(Handle, True);
end;


procedure TForm1.WMDROPFILES(var Message: TWMDROPFILES);
var
  NumFiles : longint;
  i : longint;
  buffer : array[0..255] of char;
  list : TStringList;
begin
 {How many files are being dropped}
  memo1.clear;
  list := TStringList.Create;
  list.sorted := true;

  NumFiles := DragQueryFile(Message.Drop,
                            -1,
                            nil,
                            0);
 {Accept the dropped files}
  for i := 0 to (NumFiles - 1) do begin
    DragQueryFile(Message.Drop,
                  i,
                  @buffer,
                  sizeof(buffer));
   list.Add(Buffer);
  end;

  for I := 0 to list.count - 1 do begin
   memo1.lines.Add('1. ' + list.Strings[i]);
   memo1.lines.Add('');
   StartParsingFile;
   ParseFile(list.strings[i]);
   FinishParsingFile;
   memo1.Lines.add('');
  end;
end;



procedure TForm1.FormDestroy(Sender: TObject);
begin
 ClassHeaders.Free;
 StructHeaders.Free;
 FunctionHeaders.Free;
 Functions.Free;
 ClassFunctions.Free;
end;

procedure TForm1.StartParsingFile;
begin
 Braces := 0;
 Comments := 0;
 linecomment := false;
 hashif := false;
 hashelse := false;
 hashendif := false;
 define := false;
 include := false;


 ClassHeaders.Clear;
 StructHeaders.Clear;
 FunctionHeaders.Clear;
 Functions.Clear;
 ClassFunctions.Clear;
end;

procedure TForm1.FinishParsingFile;
var
 i : integer;
begin
 // add class headers
 if (ClassHeaders.Count > 0) then begin
  memo1.lines.Add('2. Class Definitions');
  for i := 0 to ClassHeaders.Count - 1 do
   memo1.lines.Add(ClassHeaders.Strings[i]);
  memo1.lines.add('');
 end;

 // add struct headers
 if (StructHeaders.Count > 0) then begin
  memo1.lines.Add('2. Structure Definitions');
  for i := 0 to StructHeaders.Count - 1 do
   memo1.lines.Add(StructHeaders.Strings[i]);
  memo1.lines.add('');
 end;

  // add function headers
 if (FunctionHeaders.Count > 0) then begin
  memo1.lines.Add('2. Global Function Definitions');
  for i := 0 to FunctionHeaders.Count - 1 do
   memo1.lines.Add(FunctionHeaders.Strings[i]);
  memo1.lines.add('');
 end;

 // add class / struct function implementations
 if (Functions.Count > 0) then begin
  memo1.lines.Add('2. Global Function Implementations');
  for i := 0 to Functions.Count - 1 do
   memo1.lines.Add(Functions.Strings[i]);
  memo1.lines.add('');
 end;

 // add global function implementations
 if (ClassFunctions.Count > 0) then begin
  memo1.lines.Add('2. Class Function Implementations');
  for i := 0 to ClassFunctions.Count - 1 do
   memo1.lines.Add(ClassFunctions.Strings[i]);
  memo1.lines.add('');
 end;
end;

// this function takes a line of text which is a function declaration
// or implementation. It formats output to include the return type,
// the function name, the function parameters and the name of the class
// or structure it belongs to, if any.
// it returns an integer code to describe whether the line is a function
// declaration, a class function or a global function.
function TForm1.FormatFunction(text : PChar; var output : String) : Integer;
var
 pc, pc2 : PChar;
 returntype, classname, fn : String;
begin
 // first remove any unwanted crap from the line
 pc2 := text;
 pc := strpos(pc2, 'inline');
 if pc <> Nil then pc := pc + 6 else pc := pc2; // move past 'inline'
 pc2 := pc;
 pc := strpos(pc, 'extern');
 if pc <> Nil then pc := pc + 6 else pc := pc2; // move past 'extern'

 // create a special case for when there is no return type
 // (as is the case for constructors and destructors)
 // next value should be the return type.
 returntype := GetString(pc);
 classname := GetStringToColon(pc);
 if (returntype = classname) or (StrPos(PChar(returntype), '::') <> Nil) then returntype := '' else pc := pc + Length(returntype);

 // find class and function name and function parameters.
 pc2 := strpos(text, '::');
 if pc2 <> Nil then begin
  // this is a class function implementation
  result := CFI;
  pc2[0] := ' ';
  pc2[1] := ' ';
  classname := GetString(pc);
  pc := pc + Length(classname);
  fn := GetStringToBracket(pc);
  output := returntype + ', ' + fn + ', ' + classname;
 end
 else begin
  fn := GetStringToBracket(pc);
  output := returntype + ', ' + fn;
  if strpos(pc, ';') = Nil then result := GFI else result := GFH;
 end;
end;

function TForm1.GetString(text : PChar) : String;
var
 i : integer;
begin
 i := 0;
 result := '';

 repeat
  if text[i] = ' ' then inc(i) else break;
 until false;

 repeat
  result := result + text[i];
  if (text[i] = ' ') or (text[i] = #0) then break;
  inc(i);
 until false;

 Trim(result);
end;

function TForm1.GetStringtoBracket(text : PChar) : String;
var
 i : integer;
begin
 i := 0;
 result := '';

 repeat
  if text[i] = ' ' then inc(i) else break;
 until false;

 repeat
  result := result + text[i];
  if (text[i] = ')') or (text[i] = #0) then break;
  inc(i);
 until false;

 Trim(result);
end;

function TForm1.GetStringtoColon(text : PChar) : String;
var
 i : integer;
begin
 i := 0;
 result := '';

 repeat
  if text[i] = ' ' then inc(i) else break;
 until false;

 repeat
  result := result + text[i];
  if (text[i] = ':') or (text[i] = #0) then break;
  inc(i);
 until false;

 Trim(result);
end;

// read in the input file and call the parse function at the end of every
// record.
procedure TForm1.ParseFile(filename : String);
var
 F : TextFile;
 current : string;
 Delimeters : string;
 pc : Char;
begin
 AssignFile(F, filename);
 Reset(F);
 delimeters := ';{}';

 repeat
  current := '';
  repeat
   Read(F, pc);
   current := current + pc;
  until (strscan(PChar(Delimeters), pc) <> Nil) or (Eof(F) = true);
  ParseRecord(Current);
 until Eof(F) = true;

 closefile(F);
end;

function TForm1.Compare(text : string; index : integer; startstring : string) : boolean;
var
 i : integer;
begin
 result := true;
 for i := 1 to Length(Startstring) do begin
  if text[index + i - 1] <> startstring[i] then begin
   result := false;
   exit;
  end;
 end;
end;

function TForm1.RemoveLineFeeds(text : PChar) : string;
var
 i: integer;
 lastchar : char;
begin
 result := '';
 lastchar := #0;
 for i := 0 to length(text) - 1 do begin
  if (text[i] = #13) or (text[i] = #10) then continue;
  if (text[i] = ' ') and (lastchar = ' ') then continue;
  result := result + text[i];
  lastchar := text[i];
 end;
end;

end.
